/*****
      HTSensor2A.c -
      Adds hardware functionality to the framework, here the pushbuttons
      and LEDs.

			c 2019-2020 Primordial Machine Vision Systems
*****/

#include <string.h>

#include "board.h"

#include "AppTimer.h"
#include "DebugPrint.h"
#include "DebugPrintConfig.h"
#include "EventDistributor.h"
#include "SizeOf.h"

#include "ZW_application_transport_interface.h"
#include "ZW_basis_api.h"
#include "ZW_classcmd.h"
#include "ZW_system_startup_api.h"

#include "ZAF_Common_helper.h"
#include "zaf_event_helper.h"
#include "ZAF_TSE.h"
#include "ZAF_types.h"
#include "ZAF_uart_utils.h"

#include "config_app.h"


/**** Constants ****/

/* maximum number of events we can store */
#define MAXEVENT                   5

/* events that 'can wake up protocol thread' */
enum evtIDs {
	EVT_TIMER = 0,
	EVT_ZWRX,
	EVT_ZWCMD,
	EVT_APP
};

enum zwevtIDs {
	EVT_EMPTY = 0
};


/**** Globals ****/

/* number of samples left to take */
static uint8_t nsample;

/* timer to simulate measurements */
static SSwTimer simtimer;

/* FreeRTOS queue holding application events */
static QueueHandle_t evtqueue;

/* overal ZW configuration */
static SProtocolConfig_t config;

/* ZW node info setup */
static SAppNodeInfo_t nif;

/* radio setup */
static SRadioConfig_t radio;

/* security keys setup */
static uint8_t s2keys;

#ifdef DEBUGPRINT
/* buffer to hold print output - note one line! */
static uint8_t dbgbuff[96];
#endif   /* DEBUGPRINT */


/**** Prototypes ****/

static void handle_dummy(void);
static void handle_event(void);
static void simulate_ht(SSwTimer *);

static void ApplicationTask(SApplicationHandles *);


/**** Event Loop ****/

/***
    handle_event:  Process all events in the queue.  The main hardware
                   handler.
    modifies:  nsample
***/
static void handle_event(void) {
	enum zwevtIDs evt;                    /* an event in the queue */

	while (xQueueReceive(evtqueue, (void *) &evt, 0)) {
		DPRINTF("Event %d", evt);
		/* Despite BTN_EVENT_SHORT_PRESS being a macro (see board.h), it calls a
		   function so we can't use a switch statement here. */
		if (BTN_EVENT_SHORT_PRESS(BOARD_BUTTON_PB2) == (BUTTON_EVENT) evt) {
			DPRINT("  - button 2\n");
			if (0 < nsample) {
				nsample = 0;
				Board_SetLed(BOARD_LED2, LED_OFF);
			} else {
				nsample = 1;
				Board_SetLed(BOARD_LED2, LED_ON);
			}
		} else if (BTN_EVENT_SHORT_PRESS(BOARD_BUTTON_PB3) == (BUTTON_EVENT) evt) {
			DPRINT("  - button 3 = measure temperature\n");
			Board_SetLed(BOARD_LED4, LED_ON);
			TimerStart(&simtimer, 1000);
		} else if (BTN_EVENT_SHORT_PRESS(BOARD_BUTTON_PB4) == (BUTTON_EVENT) evt) {
			DPRINT("  - button 4 = measure humidity\n");
			Board_SetLed(BOARD_LED4, LED_ON);
			TimerStart(&simtimer, 1000);
		} else {
			DPRINT("  - unknown button\n");
		}
	}
}

/***
    handle_dummy:  Event handler that does nothing.  Use here for the ZWave
                   events.
***/
static void handle_dummy(void) {

	/* Do nothing. */
}

/***
    simulate_ht:  Called when measurement done.  Here, a callback on a
                  software timer that simulates the measurement.
    args:         timer - simulation timer
***/
static void simulate_ht(SSwTimer *timer) {

	Board_SetLed(BOARD_LED4, LED_OFF);
}



/**** ZWave/ZAF Application ****/

/***
    ApplicationInit:  Initialize the hardware and register the main task with
                      ZAF.  Crashes through an assert if the main task does
                      not register.
    args:             reset - reason for initializing application
    returns:   APPLICATION_RUNNING
               < 0 on failure (value depends on error)
    modifies:  config, nif, radio, s2keys, nsample
***/
ZW_APPLICATION_STATUS ApplicationInit(EResetReason_t reset) {
	int reg;                              /* true if task registered */

	AppTimerInit(EVT_TIMER, NULL);

	memset(&config, 0, sizeof(config));

	/* Per ZW_application_transport_interface.h the pointers in the structure
	   must not change although their contents can. */

	nif.DeviceOptionsMask = DEVICE_OPTIONS_MASK;
	nif.NodeType.generic = GENERIC_TYPE;
	nif.NodeType.specific = SPECIFIC_TYPE;
	nif.CommandClasses.UnSecureIncludedCC.iListLength = 0;
	nif.CommandClasses.UnSecureIncludedCC.pCommandClasses = NULL;
	nif.CommandClasses.SecureIncludedUnSecureCC.iListLength = 0;
	nif.CommandClasses.SecureIncludedUnSecureCC.pCommandClasses = NULL;
	nif.CommandClasses.SecureIncludedSecureCC.iListLength = 0;
	nif.CommandClasses.SecureIncludedSecureCC.pCommandClasses = NULL;

	radio.iListenBeforeTalkThreshold = ELISTENBEFORETALKTRESHOLD_DEFAULT;
	radio.iTxPowerLevelMax = MAX_TXPWR;
	radio.iTxPowerLevelAdjust = MEASURED_0DBM_TXPWR;
	radio.eRegion = REGION_US;

	s2keys = 0;

	config.pVirtualSlaveNodeInfoTable = NULL;
	config.pSecureKeysRequested = &s2keys;
	config.pNodeInfo = &nif;
	config.pRadioConfig = &radio;

	nsample = 0;

	Board_Init();

#ifdef DEBUGPRINT
	ZAF_UART0_enable(115200, true, false);
	DebugPrintConfig(dbgbuff, sizeof_array(dbgbuff), ZAF_UART0_tx_send);
#endif   /* DEBUGPRINT*/

	DPRINT("AppInit - registering task\n");
	reg = ZW_ApplicationRegisterTask(ApplicationTask, EVT_ZWRX, EVT_ZWCMD, &config);
	DPRINTF("          success %d\n", reg);
	ASSERT(reg);

	return APPLICATION_RUNNING;
}

/***
    ApplicationTask:  The main application task.  Loops endlessly, handling
                      events from the framework, after configuring the 
                      hardware and setting up the event queue(s).
    args:             zwsetup - ZWave configuration and queues
    modifies:  evtqueue, simtimer
***/
static void ApplicationTask(SApplicationHandles *zwsetup) {
	SEventDistributor evtdistrib;         /* event router */
	SQueueNotifying evtnotify;            /* event queue with notifications */
	StaticQueue_t evtstatic;              /* stack-allocated queue structure */
	enum zwevtIDs evtqueue_mem[MAXEVENT]; /* memory backing queue */
	EventDistributorEventHandler handlers[] = {
		AppTimerNotificationHandler,
		handle_dummy,
		handle_dummy,
		handle_event
	};                                    /* event handlers (one per evtID) */
	TaskHandle_t apptask;                 /* FreeRTOS handle for this task */

	apptask = xTaskGetCurrentTaskHandle();
	DPRINT("AppTask - start timer\n");
	AppTimerSetReceiverTask(apptask);
	DPRINT("AppTask - init ZAF\n");
	ZAF_Init(apptask, zwsetup, NULL, NULL);

	/* We never exit from this function so can use local variables rather than
	   globals, because they won't be going away. */
	DPRINT("AppTask - set up event queue\n");
	evtqueue = xQueueCreateStatic(MAXEVENT, sizeof(evtqueue_mem[0]),
	                              (uint8_t *) evtqueue_mem, &evtstatic);
	QueueNotifyingInit(&evtnotify, evtqueue, apptask, EVT_APP);
	ZAF_EventHelperInit(&evtnotify);
	EventDistributorConfig(&evtdistrib, sizeof_array(handlers), handlers, NULL);

	DPRINT("AppTask - enable buttons\n");
	Board_EnableButton(BOARD_BUTTON_PB1);
	Board_EnableButton(BOARD_BUTTON_PB2);
	Board_EnableButton(BOARD_BUTTON_PB3);
	Board_EnableButton(BOARD_BUTTON_PB4);

	DPRINT("AppTask - set up simulation timer\n");
	AppTimerRegister(&simtimer, true, simulate_ht);

	DPRINT("AppTask - set LEDs off\n");
	DPRINTF("  LED1 '%s'\n", Board_GetLedLabel(BOARD_LED1));
	Board_SetLed(BOARD_LED1, LED_OFF);
	DPRINTF("  LED2 '%s'\n", Board_GetLedLabel(BOARD_LED2));
	Board_SetLed(BOARD_LED2, LED_OFF);
	DPRINTF("  LED3 '%s'\n", Board_GetLedLabel(BOARD_LED3));
	Board_SetLed(BOARD_LED3, LED_OFF);
	DPRINTF("  LED4 '%s'\n", Board_GetLedLabel(BOARD_LED4));
	Board_SetLed(BOARD_LED4, LED_OFF);

	/*
	Board_IndicatorInit(BOARD_LED1);
	Board_IndicateStatus(BOARD_STATUS_IDLE);
	*/

	/* The FreeRTOS task loop.  Do not exit without cleaning up. */
	DPRINT("AppTask - enter main loop\n");
	for (;;) {
		EventDistributorDistribute(&evtdistrib, 10, 0);
	}
}

/***
    Transport_ApplicationCommandHandlerEx:  Incoming command class handler.
    args:                                   rxopt - receive options
                                            buff - frame buffer
                                            len - number of bytes in buffer
    returns:   RECEIVED_FRAME_STATUS_NO_SUPPORT
***/
received_frame_status_t
Transport_ApplicationCommandHandlerEx(RECEIVE_OPTIONS_TYPE_EX *rxopt,
                                      ZW_APPLICATION_TX_BUFFER *buff,
                                      uint8_t len) {

	return RECEIVED_FRAME_STATUS_NO_SUPPORT;
}
