/*****
      HTSensor2B.c -
      Adds hardware functionality to the framework, here the humidity/
      temperature sensor.

			c 2019-2020 Primordial Machine Vision Systems
*****/

#include <string.h>

#include "board.h"
#include "i2cspm.h"

#include "AppTimer.h"
#define DEBUGPRINT
#include "DebugPrint.h"
#include "DebugPrintConfig.h"
#include "EventDistributor.h"
#include "SizeOf.h"

#include "ZW_application_transport_interface.h"
#include "ZW_basis_api.h"
#include "ZW_classcmd.h"
#include "ZW_system_startup_api.h"

#include "ZAF_Common_helper.h"
#include "zaf_event_helper.h"
#include "ZAF_TSE.h"
#include "ZAF_types.h"
#include "ZAF_uart_utils.h"

#include "config_app.h"


/**** Constants ****/

/* maximum number of events we can store */
#define MAXEVENT                   5

/* number of samples in a data logging run */
#define NLOG                      50

/* time in seconds between data logs */
#define TSAMPLE                    5

/* events that 'can wake up protocol thread' */
enum evtIDs {
	EVT_TIMER = 0,
	EVT_ZWRX,
	EVT_ZWCMD,
	EVT_APP
};

enum zwevtIDs {
	EVT_EMPTY = 0
};

/* sensor commands - see Si7021 spec chapter 5 */
#define READ_H_HOLD           0xE5
#define READ_H_NOHOLD         0xF5
#define READ_T_HOLD           0xE3
#define READ_T_NOHOLD         0xF3
#define READ_T_AFTER_H        0xE0
#define TH_RESET              0xFE
#define WRITE_USER            0xE6
#define READ_USER             0xE7
#define WRITE_HEATER          0x51
#define READ_HEATER           0x11
#define READ_ID1_BYTE1        0xFA
#define READ_ID1_BYTE2        0x0F
#define READ_ID2_BYTE1        0xFC
#define READ_ID2_BYTE2        0xC9
#define READ_FW_BYTE1         0xB4
#define READ_FW_BYTE2         0xB8

/* sensor address on I2C bus - see datasheet, section 5 */
#define TH_I2CADDR            0x80


/**** Globals ****/

/* number of samples left to take */
static int nsample;

/* humidity, temperature logs */
static uint32_t log_rh[NLOG];
static int32_t log_tC[NLOG];

/* timer for LCD EXTCOMIN signal, and state flag */
static SSwTimer lcdtimer;
static uint8_t lcdon;

/* FreeRTOS queue holding application events */
static QueueHandle_t evtqueue;

/* overal ZW configuration */
static SProtocolConfig_t config;

/* ZW node info setup */
static SAppNodeInfo_t nif;

/* radio setup */
static SRadioConfig_t radio;

/* security keys setup */
static uint8_t s2keys;

#ifdef DEBUGPRINT
/* buffer to hold print output - note one line! */
static uint8_t dbgbuff[96];
#endif   /* DEBUGPRINT */


/**** Prototypes ****/

static void handle_dummy(void);
static void handle_event(void);

static void setup_i2c(void);
static const char *i2c_strerror(I2C_TransferReturn_TypeDef);
static void verify_sensor(void);
static int read_HT(uint32_t *, int32_t *);

static void setup_lcd(void);
static void toggle_extcomin(SSwTimer *);

static void ApplicationTask(SApplicationHandles *);


/**** Event Loop ****/

/***
    handle_event:  Process all events in the queue.  The main hardware
                   handler.
    modifies:  nsample
***/
static void handle_event(void) {
	enum zwevtIDs evt;                    /* an event in the queue */
	uint32_t rh;                          /* relative humidity measurement */
	int32_t tC;                           /* temperature measurement */

	while (xQueueReceive(evtqueue, (void *) &evt, 0)) {
		/* Despite BTN_EVENT_SHORT_PRESS being a macro (see board.h), it calls a
		   function so we can't use a switch statement here. */
		if (BTN_EVENT_SHORT_PRESS(BOARD_BUTTON_PB2) == (BUTTON_EVENT) evt) {
			/* The display EXTCOMIN timer will handle the sampling.  We multiple
			   by 4 because the software timer lasts 250 ms. */
			if (0 == nsample) {
				Board_SetLed(BOARD_LED2, LED_ON);
				nsample = NLOG * TSAMPLE * 4;
			}
		} else if (BTN_EVENT_SHORT_PRESS(BOARD_BUTTON_PB3) == (BUTTON_EVENT) evt) {
			Board_SetLed(BOARD_LED4, LED_ON);
			read_HT(NULL, &tC);
			Board_SetLed(BOARD_LED4, LED_OFF);
			DPRINTF("  temperature  %3d.%03d C\n", tC/1000, ((tC<0) ? -tC : tC)%1000);
		} else if (BTN_EVENT_SHORT_PRESS(BOARD_BUTTON_PB4) == (BUTTON_EVENT) evt) {
			Board_SetLed(BOARD_LED4, LED_ON);
			read_HT(&rh, NULL);
			Board_SetLed(BOARD_LED4, LED_OFF);
			DPRINTF("  humidity     %3d.%03d %%\n", rh/1000, rh%1000);
		}
	}
}

/***
    handle_dummy:  Event handler that does nothing.  Use here for the ZWave
                   events.
***/
static void handle_dummy(void) {

	/* Do nothing. */
}


/**** Sensor Interface ****/

/***
    setup_i2c:  Initialize the I2C port for the humidity/temperature sensor.
                Verify that we have a sensor (which will fail an ASSERT if
                we can't find one).
***/
static void setup_i2c(void) {
	I2CSPM_Init_TypeDef i2c0;             /* I2C setup */

	/* The values come from UG381.  Can't find a match in the hardware
	   kits for our board. */
	i2c0.port = I2C0;
	i2c0.sclPort = gpioPortC;
	i2c0.sclPin = 10;
	i2c0.portLocationScl = 14;
	i2c0.sdaPort = gpioPortC;
	i2c0.sdaPin = 11;
	i2c0.portLocationSda = 16;
	i2c0.i2cRefFreq = 0;
	i2c0.i2cMaxFreq = I2C_FREQ_STANDARD_MAX;
	i2c0.i2cClhr = i2cClockHLRStandard;
	I2CSPM_Init(&i2c0);

	verify_sensor();
}

/***
    i2c_strerror:  Convert the I2C error code to a string.
    args:          errno - result from I2C operation
    returns:   string name for error code
***/
static const char *i2c_strerror(I2C_TransferReturn_TypeDef errno) {

	switch (errno) {
	case i2cTransferInProgress:   return "in progress";
	case i2cTransferDone:         return "done";
	case i2cTransferNack:         return "NACK";
	case i2cTransferBusErr:       return "bus error";
	case i2cTransferArbLost:      return "lost arbitration";
	case i2cTransferUsageFault:   return "usage fault";
	case i2cTransferSwFault:      return "SW fault";
	default:                      return "unknown";
	}
}

/***
    verify_sensor:  Test that we can find the temp/humidity sensor.  Assumes
                    we're looking for an Si7021.
***/
static void verify_sensor(void) {
	I2C_TransferSeq_TypeDef tx;           /* packet to transfer */
	I2C_TransferReturn_TypeDef status;    /* message back from sensor */
	uint8_t txbuff[2];                    /* buffer holding outgoing message */
	uint8_t rxbuff[8];                    /* buffer holding incoming message */
	uint8_t ntry;                         /* try counter */
	int8_t found;                         /* true if we found the sensor */

	txbuff[0] = READ_ID2_BYTE1;
	txbuff[1] = READ_ID2_BYTE2;

	/* From datasheet, section 5. */
	tx.addr = TH_I2CADDR;
	tx.flags = I2C_FLAG_WRITE_READ;
	tx.buf[0].data = txbuff;
	tx.buf[0].len = sizeof_array(txbuff);
	tx.buf[1].data = rxbuff;
	tx.buf[1].len = sizeof_array(rxbuff);

	found = 0;
	/* The try cycle was used debugging the I2C setup.  Is there any reason
	   to keep it?  Technically the transfer could stall ... */
	ntry = 2;
	status = i2cTransferInProgress;
	while ((0 < ntry) && (i2cTransferInProgress == status)) {
		status = I2CSPM_Transfer(I2C0, &tx);
		if (i2cTransferDone == status) {
			/* Per spec, 0x0D is a Si7013, 0x14 a Si7020.  These aren't on our
			   devkit board. */
			if (0x15 != rxbuff[0]) {
				DPRINTF("FAILED to find sensor: ID 0x%02x not 7021 ID 0x%02x\n",
				        rxbuff[0], 0x15);
			} else {
				DPRINT("found sensor!\n");
				found = 1;
			}
			break;
		}
		ntry -= 1;
	}
	if ((0 == ntry) && (i2cTransferInProgress == status)) {
		DPRINT("FAILED to find sensor: out of tries\n");
	} else if (i2cTransferDone != status) {
		/* See em_i2c.h for error codes. */
		DPRINTF("FAILED to find sensor: error %d (%s)\n",
		        status, i2c_strerror(status));
	}

	ASSERT(found);
}

/***
    read_HT:  Grab a humidity and temperature measurement from the sensor.
              The temperature is read at the same time as the humidity.
              Pass NULL for an argument to skip that measurement.
    args:     rh - relative humidity (skip if NULL)
              tC - temperature (skip if NULL)
    returns:   0 if successful
               < 0 on failure (value depends on error)
    modifies:  rh, tC
***/
static int read_HT(uint32_t *rh, int32_t *tC) {
	I2C_TransferSeq_TypeDef tx;           /* packet to transfer */
	I2C_TransferReturn_TypeDef status;    /* message back from sensor */
	uint32_t meas;                        /* measured value */
	uint8_t txbuff[1];                    /* outgoing message */
	uint8_t rxbuff[2];                    /* incoming buffer */
	uint8_t ntry;                         /* try counter */

	if ((NULL == rh) && (NULL == tC)) {
		return 0;
	}

	tx.addr = TH_I2CADDR;
	tx.flags = I2C_FLAG_WRITE_READ;
	tx.buf[0].data = txbuff;
	tx.buf[0].len = sizeof_array(txbuff);
	tx.buf[1].data = rxbuff;
	tx.buf[1].len = sizeof_array(rxbuff);

	if (NULL != rh) {
		txbuff[0] = READ_H_HOLD;
		ntry = 2;
		status = i2cTransferInProgress;
		while ((0 < ntry) && (i2cTransferInProgress == status)) {
			status = I2CSPM_Transfer(I2C0, &tx);
			ntry -= 1;
		}

		if ((0 == ntry) && (i2cTransferInProgress == status)) {
			DPRINT("  FAILED to read humidity - out of tries\n");
			*rh = 0;
			return -1;
		} else if (i2cTransferDone != status) {
			DPRINTF("  FAILED to read humidity - error %d (%s)\n",
			        status, i2c_strerror(status));
			*rh = 0;
			return -1;
		}

		meas = ((uint32_t) rxbuff[0] << 8) + (rxbuff[1] & 0xfc);
		*rh = (meas * 125000L) >> 16;
		if (*rh < 6000) {
			*rh = 0;
		} else if (106000 < *rh) {
			*rh = 100000;
		} else {
			*rh -= 6000;
		}
	}

	if (NULL != tC) {
		if (NULL != rh) {
			txbuff[0] = READ_T_AFTER_H;
		} else {
			txbuff[0] = READ_T_HOLD;
		}

		ntry = 2;
		status = i2cTransferInProgress;
		while ((0 < ntry) && (i2cTransferInProgress == status)) {
			status = I2CSPM_Transfer(I2C0, &tx);
			ntry -= 1;
		}

		if ((0 == ntry) && (i2cTransferInProgress == status)) {
			DPRINT("  FAILED to read temperature - out of tries\n");
			*tC = 0;
			return -1;
		} else if (i2cTransferDone != status) {
			DPRINTF("  FAILED to read temperature - error %d (%s)\n",
			        status, i2c_strerror(status));
			*tC = 0;
			return -1;
		}
	
		if ((0 == ntry) && (i2cTransferInProgress == status)) {
			DPRINT("  FAILED to read temperature - out of tries\n");
			return -1;
		} else if (i2cTransferDone != status) {
			DPRINTF("  FAILED to read temperature - error %d (%s)\n",
			        status, i2c_strerror(status));
			return -1;
		}

		meas = ((uint32_t) rxbuff[0] << 8) + (rxbuff[1] & 0xfc);
		/* The formula in the spec overflows at 32 bit, ie. meas * 175720.
		   Scaling the fraction by 8 above and below. */
		*tC = (meas * 21965) >> 13;
		*tC -= 46850;
	}

	return 0;
}



/**** LCD ****/

/***
    setup_lcd:  Enabling the humidity sensor means we also enable the display.
                Set up the interface (holding the serial lines low) and
                start the refresh/de-static-ing EXTCOMIN clock.
    modifies:  lcdon
***/
static void setup_lcd(void) {

	/* Per the LCD spec SCS must be kept low while EXTCOMIN toggles for
	   a static image. */
	GPIO_PinModeSet(gpioPortD, 14, gpioModeWiredAndPullUp, 1);
	GPIO_PinOutClear(gpioPortD, 14);
	GPIO_PinModeSet(gpioPortD, 15, gpioModeWiredAndPullUp, 1);
	GPIO_PinOutSet(gpioPortD, 15);

	lcdon = 0;
	GPIO_PinModeSet(gpioPortD, 13, gpioModeWiredAndPullUp, 1);
	GPIO_PinOutClear(gpioPortD, 13);
	AppTimerRegister(&lcdtimer, true, toggle_extcomin);
	TimerStart(&lcdtimer, 250);
}

/***
    toggle_extcomin:  Change the voltage on the LCD EXTCOMIN pin.  This
                      results in a 50% duty cycle, 2 Hz signal which meets
                      the part spec (50% duty cyle, 1 s/cycle).  This also
                      handles a data logging run.
    args:             timer - on/off pulse width
    modifies:  lcdon, nsample, log_rh, log_tc
***/
static void toggle_extcomin(SSwTimer *timer) {
	int logID;                            /* which log entry to use */

	if (lcdon) {
		lcdon = 0;
		GPIO_PinOutClear(gpioPortD, 13);
	} else {
		lcdon = 1;
		GPIO_PinOutSet(gpioPortD, 13);
	}

	if (0 < nsample) {
		/* 4 here because we have four timer calls per second. */
		if (0 == (nsample % (TSAMPLE * 4))) {
			logID = (nsample / (TSAMPLE * 4)) - 1;
			read_HT(&(log_rh[logID]), &(log_tC[logID]));
			DPRINTF("log ID %-3d  rh %3d.%03d %%  tC %3d.%03d C\n", NLOG-1-logID,
			        log_rh[logID]/1000, log_rh[logID]%1000,
			        log_tC[logID]/1000, ((log_tC[logID]<0) ? -log_tC[logID] : log_tC[logID])%1000);
		}
		nsample -= 1;
		if (0 == nsample) {
			Board_SetLed(BOARD_LED2, LED_OFF);
		}
	} 
	
	TimerStart(timer, 250);
}


/**** ZWave/ZAF Application ****/

/***
    ApplicationInit:  Initialize the hardware and register the main task with
                      ZAF.  Crashes through an assert if the main task does
                      not register.
    args:             reset - reason for initializing application
    returns:   APPLICATION_RUNNING
               < 0 on failure (value depends on error)
    modifies:  config, nif, radio, s2keys, nsample
***/
ZW_APPLICATION_STATUS ApplicationInit(EResetReason_t reset) {
	int reg;                              /* true if task registered */

	AppTimerInit(EVT_TIMER, NULL);

	memset(dbgbuff, 0, sizeof_array(dbgbuff));
	memset(&config, 0, sizeof(config));

	/* Per ZW_application_transport_interface.h the pointers in the structure
	   must not change although their contents can. */

	nif.DeviceOptionsMask = DEVICE_OPTIONS_MASK;
	nif.NodeType.generic = GENERIC_TYPE;
	nif.NodeType.specific = SPECIFIC_TYPE;
	nif.CommandClasses.UnSecureIncludedCC.iListLength = 0;
	nif.CommandClasses.UnSecureIncludedCC.pCommandClasses = NULL;
	nif.CommandClasses.SecureIncludedUnSecureCC.iListLength = 0;
	nif.CommandClasses.SecureIncludedUnSecureCC.pCommandClasses = NULL;
	nif.CommandClasses.SecureIncludedSecureCC.iListLength = 0;
	nif.CommandClasses.SecureIncludedSecureCC.pCommandClasses = NULL;

	radio.iListenBeforeTalkThreshold = ELISTENBEFORETALKTRESHOLD_DEFAULT;
	radio.iTxPowerLevelMax = MAX_TXPWR;
	radio.iTxPowerLevelAdjust = MEASURED_0DBM_TXPWR;
	radio.eRegion = REGION_US;

	s2keys = 0;

	config.pVirtualSlaveNodeInfoTable = NULL;
	config.pSecureKeysRequested = &s2keys;
	config.pNodeInfo = &nif;
	config.pRadioConfig = &radio;

	nsample = 0;
	lcdon = 0;

	Board_Init();

#ifdef DEBUGPRINT
	ZAF_UART0_enable(115200, true, false);
	DebugPrintConfig(dbgbuff, sizeof_array(dbgbuff), ZAF_UART0_tx_send);
#endif   /* DEBUGPRINT*/

	reg = ZW_ApplicationRegisterTask(ApplicationTask, EVT_ZWRX, EVT_ZWCMD, &config);
	ASSERT(reg);

	return APPLICATION_RUNNING;
}

/***
    ApplicationTask:  The main application task.  Loops endlessly, handling
                      events from the framework, after configuring the 
                      hardware and setting up the event queue(s).
    args:             zwsetup - ZWave configuration and queues
    modifies:  evtqueue, lcdtimer, lcdon
***/
static void ApplicationTask(SApplicationHandles *zwsetup) {
	SEventDistributor evtdistrib;         /* event router */
	SQueueNotifying evtnotify;            /* event queue with notifications */
	StaticQueue_t evtstatic;              /* stack-allocated queue structure */
	enum zwevtIDs evtqueue_mem[MAXEVENT]; /* memory backing queue */
	EventDistributorEventHandler handlers[] = {
		AppTimerNotificationHandler,
		handle_dummy,
		handle_dummy,
		handle_event
	};		                                /* event handlers (one per evtID) */
	TaskHandle_t apptask;                 /* FreeRTOS handle for this task */

	apptask = xTaskGetCurrentTaskHandle();
	AppTimerSetReceiverTask(apptask);
	ZAF_Init(apptask, zwsetup, NULL, NULL);

	/* We never exit from this function so can use local variables rather than
	   globals, because they won't be going away. */
	evtqueue = xQueueCreateStatic(MAXEVENT, sizeof(evtqueue_mem[0]),
	                              (uint8_t *) evtqueue_mem, &evtstatic);
	QueueNotifyingInit(&evtnotify, evtqueue, apptask, EVT_APP);
	ZAF_EventHelperInit(&evtnotify);
	EventDistributorConfig(&evtdistrib, sizeof_array(handlers), handlers, NULL);

	Board_EnableButton(BOARD_BUTTON_PB1);
	Board_EnableButton(BOARD_BUTTON_PB2);
	Board_EnableButton(BOARD_BUTTON_PB3);
	Board_EnableButton(BOARD_BUTTON_PB4);

	Board_SetLed(BOARD_LED1, LED_OFF);
	Board_SetLed(BOARD_LED2, LED_OFF);
	Board_SetLed(BOARD_LED3, LED_OFF);
	Board_SetLed(BOARD_LED4, LED_OFF);

	/*
	Board_IndicatorInit(BOARD_LED1);
	Board_IndicateStatus(BOARD_STATUS_IDLE);
	*/

	setup_lcd();
	setup_i2c();
	
	/* The FreeRTOS task loop.  Do not exit without cleaning up. */
	DPRINT("AppTask - enter main loop\n");
	for (;;) {
		EventDistributorDistribute(&evtdistrib, 10, 0);
	}
}

/***
    Transport_ApplicationCommandHandlerEx:  Incoming command class handler.
    args:                                   rxopt - receive options
                                            buff - frame buffer
                                            len - number of bytes in buffer
    returns:   RECEIVED_FRAME_STATUS_NO_SUPPORT
***/
received_frame_status_t
Transport_ApplicationCommandHandlerEx(RECEIVE_OPTIONS_TYPE_EX *rxopt,
                                      ZW_APPLICATION_TX_BUFFER *buff,
                                      uint8_t len) {

	return RECEIVED_FRAME_STATUS_NO_SUPPORT;
}
