
/*****
      graphdisplay.h -
      Public interface for drawing graphs on a display.  Based on the SiLabs
      textdisplay interface.  Creates an graphing area on the display covering
      the entire width and divides the area into sub-graphs, each with a frame.

      c 2019-2020 Primordial Machine Vision Systems
*****/


#ifndef _GRAPHDISPLAY_H_
#define _GRAPHDISPLAY_H_

#include <stdbool.h>
#include "emstatus.h"

#ifdef __cplusplus
extern "C" {
#endif


/**** Constants (Internal) ****/

/* eventually this should go into emstatus.h ... */
#define GRAPHDISPLAY_EMSTATUS_BASE         (0x8014UL << 16)

/* we're not defining GRAPHDISPLAY_EMSTATUS_OK - just use EMSTATUS_OK */

/* bad parameter (usually out of range) */
#define GRAPHDISPLAY_EMSTATUS_BADPARAM     (GRAPHDISPLAY_EMSTATUS_BASE | 1)

/* no graph available */
#define GRAPHDISPLAY_EMSTATUS_NOGRAPH      (GRAPHDISPLAY_EMSTATUS_BASE | 2)

/* graph has not been set up/initialized (via _New()) */
#define GRAPHDISPLAY_EMSTATUS_NOTINIT      (GRAPHDISPLAY_EMSTATUS_BASE | 3)


/**** Structures ****/

/* External reference for the graph. */
typedef void * GRAPHDISPLAY_Handle_t;

/* Initialization structure for graphs. */
typedef struct GRAPHDISPLAY_Config_t {
	int displayDeviceNo;                  /* which display to draw on */
	uint8_t ngraph;                       /* number of graphs side by side */
	uint8_t hdraw;                        /* height drawing area (excl. frame) */
	uint8_t y0;                           /* top row display for graph+frame */
} GRAPHDISPLAY_Config_t;


/**** Public Interface ****/

/* Set up a graph on the display, incl. drawing the frame
     cfg - configuration
     handle - reference to graph area (modified)
   Returns EMSTATUS_OK or error code
*/
EMSTATUS GRAPHDISPLAY_New(GRAPHDISPLAY_Config_t *cfg,
                          GRAPHDISPLAY_Handle_t *handle);

/* Clear the graph area (incl. frame) in the display and shut down
     handle - graph area reference (modified)
   Returns EMSTATUS_OK or error code
*/
EMSTATUS GRAPHDISPLAY_Delete(GRAPHDISPLAY_Handle_t *handle);

/* Turn the pixel at (x,y) in the graph ON
     handle - graph area reference
     graphID - which sub-graph to draw  in
     x, y - coordinates within graph to draw
     on - true to draw pixel, false to erase
     redraw - true to copy change to display, false just set internally
   Returns EMSTATUS_OK or error code
*/
EMSTATUS GRAPHDISPLAY_SetPoint(GRAPHDISPLAY_Handle_t *handle,
                               uint8_t graphID, uint8_t x, uint8_t y,
                               bool on, bool redraw);

/* Clear all pixels within a graph
     handle - graph area reference
     graphID - which sub-graph to draw in
     redraw - true to copy change to display, false just set internally
   Returns EMSTATUS_OK or error code
*/
EMSTATUS GRAPHDISPLAY_Clear(GRAPHDISPLAY_Handle_t *handle, uint8_t graphID,
                            bool redraw);

/* Get the width in pixels of the drawing area (same for all sub-graphs)
     handle - graph area reference
     wdraw - width of graph (modified)
   Returns EMSTATUS_OK or error code
*/
EMSTATUS GRAPHDISPLAY_GetWidth(GRAPHDISPLAY_Handle_t *handle, uint8_t *wdraw);


#ifdef __cplusplus
}
#endif

#endif   /* _GRAPHDISPLAY_H_ */
