/*****
      HTSensor2C.c -
      Adds hardware functionality to the framework, here the display.

			c 2019-2020 Primordial Machine Vision Systems
*****/

#include <string.h>

#include "board.h"
#include "i2cspm.h"
#include "display.h"
#include "textdisplay.h"
#include "graphdisplay.h"

#include "AppTimer.h"
#define DEBUGPRINT
#include "DebugPrint.h"
#include "DebugPrintConfig.h"
#include "EventDistributor.h"
#include "SizeOf.h"

#include "ZW_application_transport_interface.h"
#include "ZW_basis_api.h"
#include "ZW_classcmd.h"
#include "ZW_system_startup_api.h"

#include "ZAF_Common_helper.h"
#include "zaf_event_helper.h"
#include "ZAF_TSE.h"
#include "ZAF_types.h"
#include "ZAF_uart_utils.h"

#include "config_app.h"


/**** Constants ****/

/* maximum number of events we can store */
#define MAXEVENT                   5

/* max number of samples in a data logging run
   (in practice, limit to width of drawing area) */
#define NLOG                      64

/* time in seconds between data logs */
#define TSAMPLE                    5

/* height of graph in pixels */
#define GRAPH_HT                  64

/* top line/row of graph on display */
#define GRAPH_TOP                 46

/* minimum range (max-min) for graph vertical interval, in milli-%/C */
/**!!!**/
/* small for test */
#define GRAPH_MINRANGE           500

/* events that 'can wake up protocol thread' */
enum evtIDs {
	EVT_TIMER = 0,
	EVT_ZWRX,
	EVT_ZWCMD,
	EVT_APP
};

enum zwevtIDs {
	EVT_EMPTY = 0
};

/* sensor commands - see Si7021 spec chapter 5 */
#define READ_H_HOLD           0xE5
#define READ_H_NOHOLD         0xF5
#define READ_T_HOLD           0xE3
#define READ_T_NOHOLD         0xF3
#define READ_T_AFTER_H        0xE0
#define TH_RESET              0xFE
#define WRITE_USER            0xE6
#define READ_USER             0xE7
#define WRITE_HEATER          0x51
#define READ_HEATER           0x11
#define READ_ID1_BYTE1        0xFA
#define READ_ID1_BYTE2        0x0F
#define READ_ID2_BYTE1        0xFC
#define READ_ID2_BYTE2        0xC9
#define READ_FW_BYTE1         0xB4
#define READ_FW_BYTE2         0xB8


/**** Structures ****/

/* sensor address on I2C bus - see datasheet, section 5 */
#define TH_I2CADDR            0x80

/* repeated samples and graph setup */
struct htsamples {
	uint8_t nsample;                      /* number samples to take */
	uint8_t tsample;                      /* time in seconds between data logs */
	uint8_t id;                           /* current sample index */
	uint32_t rh[NLOG];                    /* humidity (milli-percent) */
	uint32_t rhmin;                       /* lowest humidity in log */
	uint32_t rhmax;                       /* largest humidity in log */
	uint32_t g0min;                       /* lower humidity bound graph 0 */
	uint32_t g0scl;                       /* scaling humidity graph 0 */
	int32_t tC[NLOG];                     /* temperature (milli-degrees C) */
	int32_t tCmin;                        /* lowest temperature in log */
	int32_t tCmax;                        /* highest temperature in log */
	int32_t g1min;                        /* lower temperature bound graph 1 */
	int32_t g1scl;
};

struct appconfig {
	DISPLAY_Device_t lcd;                 /* overall display setup */
	TEXTDISPLAY_Handle_t text;            /* text half of LCD */
	GRAPHDISPLAY_Handle_t graph;          /* graphical half of LCD */
};


/**** Globals ****/

/* humidity, temperature logs */
static struct htsamples htlog;

/* setup/state of this application */
static struct appconfig app;

/* timer for sampling */
static SSwTimer sampler;

/* FreeRTOS queue holding application events */
static QueueHandle_t evtqueue;

/* overal ZW configuration */
static SProtocolConfig_t zwcfg;

/* ZW node info setup */
static SAppNodeInfo_t nif;

/* radio setup */
static SRadioConfig_t radio;

/* security keys setup */
static uint8_t s2keys;

#ifdef DEBUGPRINT
/* buffer to hold print output - note one line! */
static uint8_t dbgbuff[96];
#endif   /* DEBUGPRINT */


/**** Macros ****/

/* Return a non-zero error code (ie. not EMSTATUS_OK). */
#define RET_ON_ERR(err)                     \
	do { if (EMSTATUS_OK != (err)) { return (err); }; } while (0)


/**** Prototypes ****/

static void handle_dummy(void);
static void handle_event(void);
static void start_sampling(void);
static void sample_ht(SSwTimer *);

static void clear_log(void);
static void setup_i2c(void);
static const char *i2c_strerror(I2C_TransferReturn_TypeDef);
static void verify_sensor(void);
static uint32_t read_HT(uint32_t *, int32_t *);

static uint32_t setup_display(void);
static EMSTATUS display_humidity(uint32_t);
static EMSTATUS display_temperature(int32_t);
static void clear_graphs(void);
static EMSTATUS graph_humidity(void);
static EMSTATUS graph_temperature(void);
static EMSTATUS draw_2x2point(uint8_t, uint8_t, uint8_t, bool);

static void ApplicationTask(SApplicationHandles *);


/**** Event Loop ****/

/***
    handle_event:  Process all events in the queue.  The main hardware
                   handler.
    modifies:  
***/
static void handle_event(void) {
	enum zwevtIDs evt;                    /* an event in the queue */
	uint32_t rh;                          /* relative humidity measurement */
	int32_t tC;                           /* temperature measurement */
	uint8_t wgraph;                       /* width drawing area = max samples */
	uint32_t err;

	while (xQueueReceive(evtqueue, (void *) &evt, 0)) {
		/* Despite BTN_EVENT_SHORT_PRESS being a macro (see board.h), it calls a
		   function so we can't use a switch statement here. */
		if (BTN_EVENT_SHORT_PRESS(BOARD_BUTTON_PB2) == (BUTTON_EVENT) evt) {
			if (0 == htlog.nsample) {
				err = GRAPHDISPLAY_GetWidth(app.graph, &wgraph);
				if (EMSTATUS_OK != err) {
					return;
				}
				if (NLOG < wgraph) {
					wgraph = NLOG;
				}
				Board_SetLed(BOARD_LED2, LED_ON);
				htlog.nsample = wgraph;
				htlog.tsample = TSAMPLE;
				DPRINTF(" starting log of %u samples every %u s\n", htlog.nsample, htlog.tsample);
				start_sampling();
			} else {
				htlog.nsample = 0;
			}
		} else if (BTN_EVENT_SHORT_PRESS(BOARD_BUTTON_PB3) == (BUTTON_EVENT) evt) {
			Board_SetLed(BOARD_LED4, LED_ON);
			err = read_HT(NULL, &tC);
			if (0 == err) {
				DPRINTF("  temperature  %3d.%03d C\n", tC/1000, ((tC<0) ? -tC : tC)%1000);
				display_temperature(tC);
			}
			Board_SetLed(BOARD_LED4, LED_OFF);
		} else if (BTN_EVENT_SHORT_PRESS(BOARD_BUTTON_PB4) == (BUTTON_EVENT) evt) {
			Board_SetLed(BOARD_LED4, LED_ON);
			err = read_HT(&rh, NULL);
			if (0 == err) {
				DPRINTF("  humidity     %3d.%03d %%\n", rh/1000, rh%1000);
				display_humidity(rh);
			}
			Board_SetLed(BOARD_LED4, LED_OFF);
		}
	}
}

/***
    handle_dummy:  Event handler that does nothing.  Use here for the ZWave
                   events.
***/
static void handle_dummy(void) {

	/* Do nothing. */
}

/***
    start_sampling:  Set up the log and start the software timer.  The 
                     sampling parameters (number of samples, time between)
                     must be already set in htlog.
    modifies:  htlog
***/
static void start_sampling(void) {

	clear_log();
	clear_graphs();

	TimerStart(&sampler, htlog.tsample*1000);
}

/***
    sample_ht:  Read the humidity and temperature and store the values in
                the log.  Start another timer interval if we're not done.
    args:       timer - clock that waits until it's time for the next sample
    modifies:  htlog
***/
static void sample_ht(SSwTimer *timer) {
	uint32_t rh;                          /* humidity measurement */
	int32_t tC;                           /* temperature measurement */
	uint32_t err;

	if (htlog.id == htlog.nsample) {
		DPRINT("timer still running even though log is full\n");
		Board_SetLed(BOARD_LED2, LED_OFF);
		TimerStop(timer);
		return;
	}
	/* We signal an early stop to sampling by clearing nsample. */
	if ((0 < htlog.id) && (0 == htlog.nsample)) {
		DPRINT(" logging stopped\n");
		Board_SetLed(BOARD_LED2, LED_OFF);
		TimerStop(timer);
		return;
	}
	/* Stopped before we started. */
	if (0 == htlog.nsample) {
		return;
	}

	err = read_HT(&rh, &tC);
	if (0 != err) {
		DPRINT(" logging stopped after error reading temperature\n");
		Board_SetLed(BOARD_LED2, LED_OFF);
		TimerStop(timer);
		return;
	}

	htlog.rh[htlog.id] = rh;
	if (0 < rh) {
		if (rh < htlog.rhmin) {
			htlog.rhmin = rh;
		}
		if (htlog.rhmax < rh) {
			htlog.rhmax = rh;
		}
	}

	htlog.tC[htlog.id] = tC;
	if (0 < tC) {
		if (tC < htlog.tCmin) {
			htlog.tCmin = tC;
		}
		if (htlog.tCmax < tC) {
			htlog.tCmax = tC;
		}
	}

	err = display_humidity(rh);
	if (EMSTATUS_OK != err) {
		DPRINT(" logging stopped after display error\n");
		Board_SetLed(BOARD_LED2, LED_OFF);
		TimerStop(timer);
		return;
	}

	err = display_temperature(tC);
	if (EMSTATUS_OK != err) {
		DPRINT(" logging stopped after display error\n");
		Board_SetLed(BOARD_LED2, LED_OFF);
		TimerStop(timer);
		return;
	}

	/**!!!**/
	/* Re-work so that graph points are drawn with sending to screen, then
	   copy pixel buffer once when done, rather than twice at the end of
	   each graph_*. */
	err = graph_humidity();
	if (EMSTATUS_OK != err) {
		DPRINT(" logging stopped after graphing error\n");
		Board_SetLed(BOARD_LED2, LED_OFF);
		TimerStop(timer);
		return;
	}
	err = graph_temperature();
	if (EMSTATUS_OK != err) {
		DPRINT(" logging stopped after graphing error\n");
		Board_SetLed(BOARD_LED2, LED_OFF);
		TimerStop(timer);
		return;
	}

	htlog.id += 1;
	if (htlog.nsample <= htlog.id) {
		DPRINT(" logging done\n");
		Board_SetLed(BOARD_LED2, LED_OFF);
		TimerStop(timer);
	}
}



/**** Sensor/Log Interface ****/

/***
    clear_log:  Initialize the log.  Do not change the nsample and tsample
                members.
    modifies:  htlog
***/
static void clear_log(void) {
	uint8_t nsample;                      /* temp storage of existing nsample */
	uint8_t tsample;                      /* temp storage of existing tsample */

	/* This seems easier than zero all other members one by one. */
	nsample = htlog.nsample;
	tsample = htlog.tsample;
	memset(&htlog, 0, sizeof(htlog));
	htlog.nsample = nsample;
	htlog.tsample = tsample;

	/* Max RH is 100%.  Max storage temp is -65 to +150 C in the spec. */
	htlog.rhmin = 100001;
	htlog.tCmin = 150001;
	htlog.tCmax = -65001;
}

/***
    setup_i2c:  Initialize the I2C port for the humidity/temperature sensor.
                Verify that we have a sensor (which will fail an ASSERT if
                we can't find one).
***/
static void setup_i2c(void) {
	I2CSPM_Init_TypeDef i2c0;             /* I2C setup */

	/* Assumes the sensor is already enabled, by setting up the display 
	   before this. */

	/* The values come from UG381.  Can't find a match in the hardware
	   kits for our board. */
	i2c0.port = I2C0;
	i2c0.sclPort = gpioPortC;
	i2c0.sclPin = 10;
	i2c0.portLocationScl = 14;
	i2c0.sdaPort = gpioPortC;
	i2c0.sdaPin = 11;
	i2c0.portLocationSda = 16;
	i2c0.i2cRefFreq = 0;
	i2c0.i2cMaxFreq = I2C_FREQ_STANDARD_MAX;
	i2c0.i2cClhr = i2cClockHLRStandard;
	I2CSPM_Init(&i2c0);

	verify_sensor();
}

/***
    i2c_strerror:  Convert the I2C error code to a string.
    args:          errno - result from I2C operation
    returns:   string name for error code
***/
static const char *i2c_strerror(I2C_TransferReturn_TypeDef errno) {

	switch (errno) {
	case i2cTransferInProgress:   return "in progress";
	case i2cTransferDone:         return "done";
	case i2cTransferNack:         return "NACK";
	case i2cTransferBusErr:       return "bus error";
	case i2cTransferArbLost:      return "lost arbitration";
	case i2cTransferUsageFault:   return "usage fault";
	case i2cTransferSwFault:      return "SW fault";
	default:                      return "unknown";
	}
}

/***
    verify_sensor:  Test that we can find the temp/humidity sensor.  Assumes
                    we're looking for an Si7021.  Fails an ASSERT if sensor 
                    not found.
***/
static void verify_sensor(void) {
	I2C_TransferSeq_TypeDef tx;           /* packet to transfer */
	I2C_TransferReturn_TypeDef status;    /* message back from sensor */
	uint8_t txbuff[2];                    /* buffer holding outgoing message */
	uint8_t rxbuff[8];                    /* buffer holding incoming message */
	uint8_t ntry;                         /* try counter */
	int8_t found;                         /* true if we found the sensor */

	txbuff[0] = READ_ID2_BYTE1;
	txbuff[1] = READ_ID2_BYTE2;

	/* From datasheet, section 5. */
	tx.addr = TH_I2CADDR;
	tx.flags = I2C_FLAG_WRITE_READ;
	tx.buf[0].data = txbuff;
	tx.buf[0].len = sizeof_array(txbuff);
	tx.buf[1].data = rxbuff;
	tx.buf[1].len = sizeof_array(rxbuff);

	found = 0;
	/* The try cycle was used debugging the I2C setup.  Is there any reason
	   to keep it?  Technically the transfer could stall ... */
	ntry = 2;
	status = i2cTransferInProgress;
	while ((0 < ntry) && (i2cTransferInProgress == status)) {
		status = I2CSPM_Transfer(I2C0, &tx);
		if (i2cTransferDone == status) {
			/* Per spec, 0x0D is a Si7013, 0x14 a Si7020.  These aren't on our
			   devkit board. */
			if (0x15 != rxbuff[0]) {
				DPRINTF("FAILED to find sensor: ID 0x%02x not 7021 ID 0x%02x\n",
				        rxbuff[0], 0x15);
			} else {
				DPRINT("found sensor!\n");
				found = 1;
			}
			break;
		}
		ntry -= 1;
	}
	if ((0 == ntry) && (i2cTransferInProgress == status)) {
		DPRINT("FAILED to find sensor: out of tries\n");
	} else if (i2cTransferDone != status) {
		/* See em_i2c.h for error codes. */
		DPRINTF("FAILED to find sensor: error %d (%s)\n",
		        status, i2c_strerror(status));
	}

	ASSERT(found);
}

/***
    read_HT:  Grab a humidity and temperature measurement from the sensor.
              The temperature is read at the same time as the humidity.
              Pass NULL for an argument to skip that measurement.
    args:     rh - relative humidity (skip if NULL)
              tC - temperature (skip if NULL)
    returns:   0 if successful, > 0 on error
    modifies:  rh, tC
***/
static uint32_t read_HT(uint32_t *rh, int32_t *tC) {
	I2C_TransferSeq_TypeDef tx;           /* packet to transfer */
	I2C_TransferReturn_TypeDef status;    /* message back from sensor */
	uint32_t meas;                        /* measured value */
	uint8_t txbuff[1];                    /* outgoing message */
	uint8_t rxbuff[2];                    /* incoming buffer */
	uint8_t ntry;                         /* try counter */

	if ((NULL == rh) && (NULL == tC)) {
		return 0;
	}

	tx.addr = TH_I2CADDR;
	tx.flags = I2C_FLAG_WRITE_READ;
	tx.buf[0].data = txbuff;
	tx.buf[0].len = sizeof_array(txbuff);
	tx.buf[1].data = rxbuff;
	tx.buf[1].len = sizeof_array(rxbuff);

	if (NULL != rh) {
		txbuff[0] = READ_H_HOLD;
		ntry = 2;
		status = i2cTransferInProgress;
		while ((0 < ntry) && (i2cTransferInProgress == status)) {
			status = I2CSPM_Transfer(I2C0, &tx);
			ntry -= 1;
		}

		if ((0 == ntry) && (i2cTransferInProgress == status)) {
			DPRINT("  FAILED to read humidity - out of tries\n");
			*rh = 0;
			return 1;
		} else if (i2cTransferDone != status) {
			DPRINTF("  FAILED to read humidity - error %d (%s)\n",
			        status, i2c_strerror(status));
			*rh = 0;
			return 1;
		}

		meas = ((uint32_t) rxbuff[0] << 8) + (rxbuff[1] & 0xfc);
		*rh = (meas * 125000L) >> 16;
		if (*rh < 6000) {
			*rh = 0;
		} else if (106000 < *rh) {
			*rh = 100000;
		} else {
			*rh -= 6000;
		}
	}

	if (NULL != tC) {
		if (NULL != rh) {
			txbuff[0] = READ_T_AFTER_H;
		} else {
			txbuff[0] = READ_T_HOLD;
		}

		ntry = 2;
		status = i2cTransferInProgress;
		while ((0 < ntry) && (i2cTransferInProgress == status)) {
			status = I2CSPM_Transfer(I2C0, &tx);
			ntry -= 1;
		}

		if ((0 == ntry) && (i2cTransferInProgress == status)) {
			DPRINT("  FAILED to read temperature - out of tries\n");
			*tC = 0;
			return 1;
		} else if (i2cTransferDone != status) {
			DPRINTF("  FAILED to read temperature - error %d (%s)\n",
			        status, i2c_strerror(status));
			*tC = 0;
			return 1;
		}
	
		if ((0 == ntry) && (i2cTransferInProgress == status)) {
			DPRINT("  FAILED to read temperature - out of tries\n");
			return 1;
		} else if (i2cTransferDone != status) {
			DPRINTF("  FAILED to read temperature - error %d (%s)\n",
			        status, i2c_strerror(status));
			return 1;
		}

		meas = ((uint32_t) rxbuff[0] << 8) + (rxbuff[1] & 0xfc);
		/* The formula in the spec overflows at 32 bit, ie. meas * 175720.
		   Scaling the fraction by 8 above and below. */
		*tC = (meas * 21965) >> 13;
		*tC -= 46850;
	}

	return 0;
}


/**** LCD ****/

/***
    setup_display:  Prepare the display, including text support.  Will stop
                    halfway through process if any step fails.  As a side
                    effect of initializing the LCD, it's also enabled - which
                    jointly enables the sensor.
    returns:   0 if successful, > 0 on error (value depends on error)
    modifies:  app
***/
static uint32_t setup_display(void) {
	TEXTDISPLAY_Config_t textcfg;         /* textdisplay configuration */
	GRAPHDISPLAY_Config_t graphcfg;       /* graphdisplay configuration */
	uint32_t err;

	err = DISPLAY_Init();
	RET_ON_ERR(err);

	err = DISPLAY_DeviceGet(0, &app.lcd);
	RET_ON_ERR(err);

	/* Function only uses these members of cfg: displayDeviceNo, 
	   scrollEnable, lfToCrLf. */
	textcfg.displayDeviceNo = 0;
	textcfg.scrollEnable = 0;
	textcfg.lfToCrLf = 0;
	textcfg.maxlines = 2;
	err = TEXTDISPLAY_New(&textcfg, &app.text);
	RET_ON_ERR(err);

	graphcfg.displayDeviceNo = 0;
	graphcfg.y0 = GRAPH_TOP;
	graphcfg.hdraw = GRAPH_HT;
	graphcfg.ngraph = 2;
	err = GRAPHDISPLAY_New(&graphcfg, &app.graph);
	RET_ON_ERR(err);

	return 0;
}

/***
    display_humidity:  Print the humidity (XXX.X %) to the first line of
                       the text display.  The rest of the display isn't
                       changed.
    args:              rh - humidity (milli-percent)
    returns:   0 if successful, > 0 on error (value depends on error)
***/
static EMSTATUS display_humidity(uint32_t rh) {
	uint32_t pwr10;                       /* which digit to print */
	EMSTATUS err;

	/* Maximum width is 7 characters, but since we should never see 100%,
	   the first digit will be a blank.  Therefore left-justify the
	   string so it looks centered, usually. */

	err = TEXTDISPLAY_WriteString(app.text, TEXTDISPLAY_ESC_SEQ_CURSOR_HOME_VT52);
	RET_ON_ERR(err);

	/* Round up the tenths digit. */
	rh += 50;

	pwr10 = 100000;
	while (10 < pwr10) {
		/* Don't print leading zeros before the unit's digit (pwr10==1000),
		   put do print it and after the decimal point. */
		if ((0 == (rh / pwr10)) && (1000 < pwr10)) {
			err = TEXTDISPLAY_WriteChar(app.text, ' ');
			RET_ON_ERR(err);
		} else {
			err = TEXTDISPLAY_WriteChar(app.text, '0' + (rh / pwr10));
			RET_ON_ERR(err);
		}
		if (1000 == pwr10) {
			err = TEXTDISPLAY_WriteChar(app.text, '.');
			RET_ON_ERR(err);
		}

		rh %= pwr10;
		pwr10 /= 10;
	}

	err = TEXTDISPLAY_WriteChar(app.text, ' ');
	RET_ON_ERR(err);

	err = TEXTDISPLAY_WriteChar(app.text, '/');
	RET_ON_ERR(err);

	return EMSTATUS_OK;
}

/***
    display_temperature:  Print the temperature ([-]XXX.X degC) to the 
                          second line of the text display.  The rest of
                          the display isn't changed.
    args:                 tC - temperature (milli-degree C)
    returns:   0 if successful, > 0 on error (value depends on error)
***/
static EMSTATUS display_temperature(int32_t tC) {
	uint32_t pwr10;                       /* which digit to print */
	EMSTATUS err;

	/* Maximum width is 7 characters, either 'XXX.X C' or '-XX.X C'.  As
	   with humidity, the first character will usually not print so put
	   it in the first column, apparently centering the value. */

	err = TEXTDISPLAY_WriteString(app.text, TEXTDISPLAY_ESC_SEQ_CURSOR_HOME_VT52);
	RET_ON_ERR(err);
	err = TEXTDISPLAY_WriteString(app.text, TEXTDISPLAY_ESC_SEQ_CURSOR_DOWN_ONE_LINE);
	RET_ON_ERR(err);

	if (tC < 0) {
		err = TEXTDISPLAY_WriteChar(app.text, '-');
		RET_ON_ERR(err);
		tC = -tC;
		/* Temp storage spec is -65 C, so don't worry about -100.  The minus
		   sign has replaced the 100's digit. */
		pwr10 = 10000;
	} else {
		pwr10 = 100000;
	}

	/* Round up the tenths digit. */
	tC += 50;

	while (10 < pwr10) {
		/* Don't print leading zeros before the unit's digit (pwr10==1000),
		   put do print it and after the decimal point. */
		if ((0 == (tC / pwr10)) && (1000 < pwr10)) {
			err = TEXTDISPLAY_WriteChar(app.text, ' ');
			RET_ON_ERR(err);
		} else {
			err = TEXTDISPLAY_WriteChar(app.text, '0' + (tC / pwr10));
			RET_ON_ERR(err);
		}
		if (1000 == pwr10) {
			err = TEXTDISPLAY_WriteChar(app.text, '.');
			RET_ON_ERR(err);
		}

		tC %= pwr10;
		pwr10 /= 10;
	}

	err = TEXTDISPLAY_WriteChar(app.text, ' ');
	RET_ON_ERR(err);

	err = TEXTDISPLAY_WriteChar(app.text, ';');
	RET_ON_ERR(err);

	return EMSTATUS_OK;
}

/***
    clear_graphs:  Erase any points already drawn in both sub-graphs.
***/
static void clear_graphs(void) {

	GRAPHDISPLAY_Clear(app.graph, 0, false);
	GRAPHDISPLAY_Clear(app.graph, 1, true);
}

/***
    graph_humidity:  Draw the last measurement in the log to the graph.
                     May re-scale and re-draw the entire log if necessary.
                     The log ID must point to the last sample.
    returns:   0 if successful, > 0 on error (value depends on error)
    modifies:  htlog
***/
static EMSTATUS graph_humidity(void) {
	uint8_t wgraph;                       /* sub-graph width */
	uint8_t x,y;                          /* pixel coordinates in sub-graph */
	EMSTATUS err;

	GRAPHDISPLAY_GetWidth(app.graph, &wgraph);
	if (wgraph <= htlog.id) {
		DPRINTF("too many points for graph (%d vs. w %d), need scrolling\n",
		        htlog.id, wgraph);
		return 1;
	}

	if (0 == htlog.rh[htlog.id]) {
		return 0;
	}

	/* If graph bounds are correct then just draw the point.  Otherwise
	   must re-calculate the bounds, clear the graph, and draw all points. */
	if ((0 < htlog.g0min) && (htlog.g0min < htlog.rhmin) &&
	    (0 < htlog.g0scl) &&
	    ((htlog.rhmax - htlog.g0min) < (GRAPH_HT * htlog.g0scl))) {
		y = (((htlog.rh[htlog.id] - htlog.g0min) / 100) * GRAPH_HT) / htlog.g0scl;
		err = draw_2x2point(0, htlog.id, y, true);
		RET_ON_ERR(err);
	} else {
		if ((htlog.rhmax - htlog.rhmin) < GRAPH_HT) {
			/* Center range (no scaling) about average. */
			htlog.g0min = (((htlog.rhmin + htlog.rhmax - GRAPH_MINRANGE) / 2) / 100) * 100;
			htlog.g0scl = GRAPH_MINRANGE / GRAPH_HT;
		} else {
			htlog.g0min = (htlog.rhmin / 100) * 100;
			htlog.g0scl =
				((htlog.rhmax - htlog.g0min) + GRAPH_HT - 1) / GRAPH_HT;
			err = GRAPHDISPLAY_Clear(app.graph, 0, false);
			RET_ON_ERR(err);
			for (x=0; x<htlog.id; x++) {
				if (0 == htlog.rh[x]) {
					continue;
				}
				y = (((htlog.rh[x] - htlog.g0min) / 100) * GRAPH_HT) / htlog.g0scl;
				err = draw_2x2point(0, x, y, false);
				RET_ON_ERR(err);
			}
		}

		y = (((htlog.rh[htlog.id] - htlog.g0min) / 100) * GRAPH_HT) / htlog.g0scl;
		err = draw_2x2point(0, htlog.id, y, true);
		RET_ON_ERR(err);
	}

	return EMSTATUS_OK;
}

/***
    graph_temperature:  Draw the last measurement in the log to the graph.
                        May re-scale and re-draw the entire log if necessary.
                        The log ID must point to the last sample.
    returns:   0 if successful, > 0 on error (value depends on error)
    modifies:  htlog
***/
static EMSTATUS graph_temperature(void) {
	uint8_t wgraph;                       /* sub-graph width */
	uint8_t x,y;                          /* pixel coordinates in sub-graph */
	EMSTATUS err;

	GRAPHDISPLAY_GetWidth(app.graph, &wgraph);
	if (wgraph <= htlog.id) {
		DPRINTF("too many points for graph (%d vs. width %d), need scrolling\n",
		        htlog.id, wgraph);
		return 1;
	}

	if (0 == htlog.tC[htlog.id]) {
		return 0;
	}

	/* If graph bounds are correct then just draw the point.  Otherwise
	   must re-calculate the bounds, clear the graph, and draw all points. */
	if ((0 < htlog.g1min) && (htlog.g1min < htlog.tCmin) &&
	    (0 < htlog.g1scl) &&
	    ((htlog.tCmax - htlog.g1min) < (GRAPH_HT * htlog.g1scl))) {
		y = (((htlog.tC[htlog.id] - htlog.g1min) / 100) * GRAPH_HT) / htlog.g1scl;
		err = draw_2x2point(1, htlog.id, y, true);
		RET_ON_ERR(err);
	} else {
		if ((htlog.tCmax - htlog.tCmin) < GRAPH_HT) {
			/* Center range (no scaling) about average. */
			htlog.g1min = (((htlog.tCmin + htlog.tCmax - GRAPH_MINRANGE) / 2) / 100) * 100;
			htlog.g1scl = GRAPH_MINRANGE / GRAPH_HT;
		} else {
			htlog.g1min = (htlog.tCmin / 100) * 100;
			htlog.g1scl =
				((htlog.tCmax - htlog.g1min) + GRAPH_HT - 1) / GRAPH_HT;
			err = GRAPHDISPLAY_Clear(app.graph, 1, false);
			RET_ON_ERR(err);
			for (x=0; x<htlog.id; x++) {
				if (0 == htlog.tC[x]) {
					continue;
				}
				y = (((htlog.tC[x] - htlog.g1min) / 100) * GRAPH_HT) / htlog.g1scl;
				err = draw_2x2point(1, x, y, false);
				RET_ON_ERR(err);
			}
		}

		y = (((htlog.tC[htlog.id] - htlog.g1min) / 100) * GRAPH_HT) / htlog.g1scl;
		err = draw_2x2point(1, htlog.id, y, true);
		RET_ON_ERR(err);
	}

	return EMSTATUS_OK;
}

/***
    draw_2x2point:  
                    
    args:           
                    
    returns:   0 if successful
               < 0 on failure (value depends on error)
    modifies:  
***/
static EMSTATUS draw_2x2point(uint8_t graphID, uint8_t x, uint8_t y, bool redraw) {
	EMSTATUS err;

	/* We don't check if x, y legit because the graph routine does that.
	   Ignore OOB (ie. BADPARAM) errors. */

	err = GRAPHDISPLAY_SetPoint(app.graph, graphID, x, y, true, redraw);
	if ((EMSTATUS_OK != err) && (GRAPHDISPLAY_EMSTATUS_BADPARAM != err)) {
		return err;
	}

	err = GRAPHDISPLAY_SetPoint(app.graph, graphID, x-1, y, true, redraw);
	if ((EMSTATUS_OK != err) && (GRAPHDISPLAY_EMSTATUS_BADPARAM != err)) {
		return err;
	}

	err = GRAPHDISPLAY_SetPoint(app.graph, graphID, x, y-1, true, redraw);
	if ((EMSTATUS_OK != err) && (GRAPHDISPLAY_EMSTATUS_BADPARAM != err)) {
		return err;
	}

	err = GRAPHDISPLAY_SetPoint(app.graph, graphID, x-1, y-1, true, redraw);
	if ((EMSTATUS_OK != err) && (GRAPHDISPLAY_EMSTATUS_BADPARAM != err)) {
		return err;
	}

	return EMSTATUS_OK;
}



/**** ZWave/ZAF Application ****/

/***
    ApplicationInit:  Initialize the hardware and register the main task with
                      ZAF.  Crashes through an assert if the main task does
                      not register.
    args:             reset - reason for initializing application
    returns:   APPLICATION_RUNNING
               < 0 on failure (value depends on error)
    modifies:  zwcfg, nif, radio, s2keys, nsample
***/
ZW_APPLICATION_STATUS ApplicationInit(EResetReason_t reset) {
	uint32_t err;

	AppTimerInit(EVT_TIMER, NULL);

	memset(dbgbuff, 0, sizeof_array(dbgbuff));
	memset(&zwcfg, 0, sizeof(zwcfg));

	/* Per ZW_application_transport_interface.h the pointers in the structure
	   must not change although their contents can. */

	nif.DeviceOptionsMask = DEVICE_OPTIONS_MASK;
	nif.NodeType.generic = GENERIC_TYPE;
	nif.NodeType.specific = SPECIFIC_TYPE;
	nif.CommandClasses.UnSecureIncludedCC.iListLength = 0;
	nif.CommandClasses.UnSecureIncludedCC.pCommandClasses = NULL;
	nif.CommandClasses.SecureIncludedUnSecureCC.iListLength = 0;
	nif.CommandClasses.SecureIncludedUnSecureCC.pCommandClasses = NULL;
	nif.CommandClasses.SecureIncludedSecureCC.iListLength = 0;
	nif.CommandClasses.SecureIncludedSecureCC.pCommandClasses = NULL;

	radio.iListenBeforeTalkThreshold = ELISTENBEFORETALKTRESHOLD_DEFAULT;
	radio.iTxPowerLevelMax = MAX_TXPWR;
	radio.iTxPowerLevelAdjust = MEASURED_0DBM_TXPWR;
	radio.eRegion = REGION_US;

	s2keys = 0;

	zwcfg.pVirtualSlaveNodeInfoTable = NULL;
	zwcfg.pSecureKeysRequested = &s2keys;
	zwcfg.pNodeInfo = &nif;
	zwcfg.pRadioConfig = &radio;

	clear_log();

	err = Board_Init();
	ASSERT(0 == err);
	err = setup_display();
	ASSERT(0 == err);

#ifdef DEBUGPRINT
	ZAF_UART0_enable(115200, true, false);
	DebugPrintConfig(dbgbuff, sizeof_array(dbgbuff), ZAF_UART0_tx_send);
	DPRINT("\n");
#endif   /* DEBUGPRINT*/

	err = ZW_ApplicationRegisterTask(ApplicationTask, EVT_ZWRX, EVT_ZWCMD, &zwcfg);
	ASSERT(err);

	return APPLICATION_RUNNING;
}

/***
    ApplicationTask:  The main application task.  Loops endlessly, handling
                      events from the framework, after configuring the 
                      hardware and setting up the event queue(s).
    args:             zwsetup - ZWave configuration and queues
    modifies:  evtqueue
***/
static void ApplicationTask(SApplicationHandles *zwsetup) {
	SEventDistributor evtdistrib;         /* event router */
	SQueueNotifying evtnotify;            /* event queue with notifications */
	StaticQueue_t evtstatic;              /* stack-allocated queue structure */
	enum zwevtIDs evtqueue_mem[MAXEVENT]; /* memory backing queue */
	EventDistributorEventHandler handlers[] = {
		AppTimerNotificationHandler,
		handle_dummy,
		handle_dummy,
		handle_event
	};		                                /* event handlers (one per evtID) */
	TaskHandle_t apptask;                 /* FreeRTOS handle for this task */

	apptask = xTaskGetCurrentTaskHandle();
	AppTimerSetReceiverTask(apptask);
	ZAF_Init(apptask, zwsetup, NULL, NULL);

	/* We never exit from this function so can use local variables rather than
	   globals, because they won't be going away. */
	evtqueue = xQueueCreateStatic(MAXEVENT, sizeof(evtqueue_mem[0]),
	                              (uint8_t *) evtqueue_mem, &evtstatic);
	QueueNotifyingInit(&evtnotify, evtqueue, apptask, EVT_APP);
	ZAF_EventHelperInit(&evtnotify);
	EventDistributorConfig(&evtdistrib, sizeof_array(handlers), handlers, NULL);

	Board_EnableButton(BOARD_BUTTON_PB1);
	Board_EnableButton(BOARD_BUTTON_PB2);
	Board_EnableButton(BOARD_BUTTON_PB3);
	Board_EnableButton(BOARD_BUTTON_PB4);

	Board_SetLed(BOARD_LED1, LED_OFF);
	Board_SetLed(BOARD_LED2, LED_OFF);
	Board_SetLed(BOARD_LED3, LED_OFF);
	Board_SetLed(BOARD_LED4, LED_OFF);

	/*
	Board_IndicatorInit(BOARD_LED1);
	Board_IndicateStatus(BOARD_STATUS_IDLE);
	*/

	setup_i2c();
	AppTimerRegister(&sampler, true, sample_ht);

	/* The FreeRTOS task loop.  Do not exit without cleaning up. */
	DPRINT("AppTask - enter main loop\n");
	for (;;) {
		EventDistributorDistribute(&evtdistrib, 10, 0);
	}
}

/***
    Transport_ApplicationCommandHandlerEx:  Incoming command class handler.
    args:                                   rxopt - receive options
                                            buff - frame buffer
                                            len - number of bytes in buffer
    returns:   RECEIVED_FRAME_STATUS_NO_SUPPORT
***/
received_frame_status_t
Transport_ApplicationCommandHandlerEx(RECEIVE_OPTIONS_TYPE_EX *rxopt,
                                      ZW_APPLICATION_TX_BUFFER *buff,
                                      uint8_t len) {

	return RECEIVED_FRAME_STATUS_NO_SUPPORT;
}
