
/*****
      graphdisplay.c -
      Draw graphs on a display.  Reserves rows in the display, and divides
      the whole graph area into multiple sub-graphs (IDs start at 0).  Each 
      has a frame.  Set individual pixels within a sub-graph ON or OFF.

      c 2019-2020 Primordial Machine Vision Systems
*****/

#include <stdio.h>
#include <stdbool.h>
#include <string.h>

#include "display.h"
#include "graphdisplay.h"


/**** Constants ****/

/* width/height of tics in pixels */
#define TICSIZE                  4

/* draw tics this many pixels apart */
#define TICSEP                  10

/* white space around graph in pixels */
#define WGAP                     2

/* width of frame border in pixels */
#define WBORDER                  2


/**** Structures ****/

/* Setup of graph on display.  Dimensions are in terms of pixels, not the
   compact bit-packing of the buffer (buffer size is ncol/8 x nrow). */
typedef struct GRAPHDISPLAY_Device_t {
	int dispID;                           /* which display we're drawing on */
	DISPLAY_Device_t display;             /* display properties */
	uint8_t nrow;                         /* number lines/rows on display */
	uint8_t ncol;                         /* number columns/width on display */
	uint8_t y0;                           /* first row in display for area */
	uint8_t ngraph;                       /* number sub-graphs */
	uint8_t hdraw;                        /* height drawing area (w/o frames) */
	uint8_t wdraw;                        /* width drawing area (w/o frames) */
	uint8_t wgraph;                       /* width of drawing area + frame */
	DISPLAY_PixelMatrix_t buff;           /* memory buffer for entire graph */
	bool init;                            /* true when initialized */
} GRAPHDISPLAY_Device_t;


/**** Globals ****/

/* allow one graph per display device */
/*   would need a GRAPHDISPLAY_DEVICES_MAX in the kit config */
static GRAPHDISPLAY_Device_t graphtable[DISPLAY_DEVICES_MAX] = { 0 };


/**** Macros ****/

/* Return a non-zero error code (ie. not EMSTATUS_OK). */
#define RET_ON_ERR(err)                                                 \
	do { if (EMSTATUS_OK != (err)) { return (err); }; } while (0)

/* Evaluates to the index in a pixel buffer of width ncol for point (x,y).
   Inputs are in pixels, output is the buffer index (bytes). */
#define XYTOIND(x, y, ncol)                                             \
	((((y) * (uint32_t) (ncol)) + (x)) / 8)

/* Evaluates to the left pixel column of the drawing area inside the frame
   for sub-graph subID.  g is the graph setup (GRAPHDISPLAY_Device_t). */
#define X0(g, subID)                                                    \
	((0 == (subID)) ?                                                     \
	  (TICSIZE + WBORDER + WGAP) :                                        \
	  ((g->ngraph - 1) == subID) ?                                        \
	    (((subID) * g->wgraph) + (2 * WGAP)) :                            \
	    ((subID) * g->wgraph) + (2 * WGAP) + (TICSIZE/2))

/* Evaluates to the bottom pixel row of the drawing area inside the frame.
   The vertical axis is of course inverted, ie. subtract y from this. */
#define Y0(g)                                                           \
	(g->hdraw + WBORDER + WGAP)

/* Set/clear the bit in the buffer b corresponding to the (x,y) pixel.  ncol
   is the width of the buffer in pixels (bits, not bytes).  Pixels are
   inverted, ie. bit 0 is dark on LCD. */
#define SET_PIXEL(b, x, y, ncol)                                        \
	do { (b)[XYTOIND(x,y,ncol)] &= ~(1 << ((x) % 8)); } while (0)
#define CLEAR_PIXEL(b, x, y, ncol)                                      \
	do { (b)[XYTOIND(x,y,ncol)] |= 1 << ((x) % 8); } while (0)


/**** Prototypes ****/

static EMSTATUS draw_frames(GRAPHDISPLAY_Device_t *);


/**** Public Functions ****/

/***
    GRAPHDISPLAY_New:  Create a graph display on a device and draw frames
                       for it to the device.
    args:              cfg - setup of graph area
                       handle - reference to graph area
    returns:   0 if successful, > 0 on error (EMSTATUS code)
    modifies:  handle
***/
EMSTATUS GRAPHDISPLAY_New(GRAPHDISPLAY_Config_t *cfg,
                          GRAPHDISPLAY_Handle_t *handle) {
	GRAPHDISPLAY_Device_t *graph;         /* graph data */
	uint8_t tblID;                        /* which graph table entry to use */
	uint8_t wframe;                       /* width (or height) of frame */
	uint32_t err;

	for (tblID=0; (tblID<DISPLAY_DEVICES_MAX) && graphtable[tblID].init; tblID++) {
		/* do nothing - just test initialization */
	}
	if (DISPLAY_DEVICES_MAX == tblID) {
		return GRAPHDISPLAY_EMSTATUS_NOGRAPH;
	}
	graph = &graphtable[tblID];

	err = DISPLAY_DeviceGet(cfg->displayDeviceNo, &graph->display);
	RET_ON_ERR(err);

	wframe = 2 * (WGAP + WBORDER) + TICSIZE;
	graph->y0 = cfg->y0;
	graph->hdraw = cfg->hdraw;
	graph->nrow = graph->hdraw + wframe;
	if (graph->display.geometry.height < (graph->y0 + graph->nrow)) {
		return GRAPHDISPLAY_EMSTATUS_BADPARAM;
	}

	graph->ngraph = cfg->ngraph;
	graph->ncol = graph->display.geometry.width;
	if (graph->ncol < (graph->ngraph * wframe)) {
		return GRAPHDISPLAY_EMSTATUS_BADPARAM;
	}
	graph->wdraw = (graph->ncol - (graph->ngraph * wframe)) / graph->ngraph;
	graph->wgraph = graph->wdraw + wframe;

	/* The comments in display.c say that the stride must be at least 160 
	   pixels with EMWIN_WORKAROUND, but textdisplay.c uses the same width
	   with it as without.  Is the workaround needed any more?  If so then
	   adjust the wrapped ncol.  The SDK doesn't include EMWIN (the gecko
	   SDK does), so this isn't an issue for us. */
	err = graph->display.pPixelMatrixAllocate(&graph->display, graph->ncol,
#ifdef EMWIN_WORKAROUND
	                                          graph->ncol,
#endif   /* EMWIN_WORKAROUND */
	                                          graph->nrow, &graph->buff);
	RET_ON_ERR(err);

	err = draw_frames(graph);
	RET_ON_ERR(err);
	err = graph->display.pPixelMatrixDraw(&graph->display, graph->buff,
	                                      0, graph->ncol,
#ifdef EMWIN_WORKAROUND
	                                      graph->ncol,
#endif   /* EMWIN_WORKAROUND */
	                                      graph->y0, graph->nrow);
	RET_ON_ERR(err);

	graph->init = true;

	*handle = (GRAPHDISPLAY_Handle_t *) graph;
	return EMSTATUS_OK;
}

/***
    GRAPHDISPLAY_Delete:  Clean up any resources we've allocated internally.  
                          The handle is set to NULL when done.
    args:                 handle - reference to graph area
    returns:   0 if successful, > 0 on error (EMSTATUS code)
    modifies:  handle
***/
EMSTATUS GRAPHDISPLAY_Delete(GRAPHDISPLAY_Handle_t *handle) {
	GRAPHDISPLAY_Device_t *graph;         /* graph data */
	uint32_t err;

	graph = (GRAPHDISPLAY_Device_t *) handle;

	if ((NULL == graph) || !graph->init) {
		return GRAPHDISPLAY_EMSTATUS_NOTINIT;
	}

	if (graph->buff) {
		err = graph->display.pPixelMatrixFree(&graph->display, graph->buff);
	}
	memset(graph, 0, sizeof(*graph));

	*handle = NULL;
	return EMSTATUS_OK;
}

/***
    GRAPHDISPLAY_SetPoint:  Draw/erase a pixel in a sub-graph.  Update the
                            display if required, otherwise just keep internally
                            (if you're redrawing a complete graph, only do an
                            update for the last pixel you add).
    args:                   handle - reference to graph area
                            graphID - which sub-graph to draw in
                            x, y - coordinates within graph to draw
                            on - true to draw pixel, false to erase
                            redraw - true to copy change to display
    returns:   0 if successful, > 0 on error (EMSTATUS code)
    modifies:  
***/
EMSTATUS GRAPHDISPLAY_SetPoint(GRAPHDISPLAY_Handle_t *handle, uint8_t graphID,
                               uint8_t x, uint8_t y, bool on, bool redraw) {
	GRAPHDISPLAY_Device_t *graph;         /* graph data */
	uint8_t xpt, ypt;                     /* point coordinates inside frame */
	uint32_t err;

	graph = (GRAPHDISPLAY_Device_t *) handle;

	if ((NULL == graph) || !graph->init) {
		return GRAPHDISPLAY_EMSTATUS_NOTINIT;
	}

	if ((graph->wdraw <= x) || (graph->hdraw <= y)) {
		return GRAPHDISPLAY_EMSTATUS_BADPARAM;
	}
	xpt = X0(graph, graphID) + x;
	ypt = Y0(graph) - y;

	if (on) {
		SET_PIXEL((uint8_t *) graph->buff, xpt, ypt, graph->ncol);
	} else {
		CLEAR_PIXEL((uint8_t *) graph->buff, xpt, ypt, graph->ncol);
	}

	if (redraw) {
		err = graph->display.pPixelMatrixDraw(&graph->display, graph->buff,
		                                      0, graph->ncol,
#ifdef EMWIN_WORKAROUND
		                                      graph->ncol,
#endif   /* EMWIN_WORKAROUND */
		                                      graph->y0, graph->nrow);
		RET_ON_ERR(err);
	}

	return EMSTATUS_OK;
}

/***
    GRAPHDISPLAY_Clear:  Erase all pixels inside the sub-graph.  Does not
                         touch the frame.
    args:                handle - reference to graph area
                         subID - which sub-graph to wipe
                         redraw - true to put cleared array on display
    returns:   0 if successful, > 0 on error (EMSTATUS code)
    modifies:  graph buff contents
***/
EMSTATUS GRAPHDISPLAY_Clear(GRAPHDISPLAY_Handle_t *handle, uint8_t subID,
                            bool redraw) {
	GRAPHDISPLAY_Device_t *graph;         /* graph data */
	uint8_t *buff;                        /* pixel buffer */
	uint8_t xst, xend;                    /* horizontal line bounds (incl.) */
	uint8_t yst, yend;                    /* vertical line bounds (incl.) */
	uint8_t x, y;                         /* pixel coordinates */
	uint32_t xy;                          /* pixel index in buffer */
	uint32_t err;

	graph = (GRAPHDISPLAY_Device_t *) handle;
	buff = (uint8_t *) graph->buff;

	if ((NULL == graph) || (!graph->init)) {
		return GRAPHDISPLAY_EMSTATUS_NOTINIT;
	}

	/* +1/-1 for inclusive end. */
	xst = X0(graph, subID);
	xend = xst + graph->wdraw - 1;
	yend = Y0(graph);
	yst = yend - graph->hdraw + 1;
	for (y=yst; y<=yend; y++) {
		for (x=xst; x<=xend; x++) {
			if ((0 == (x % 8)) && ((x + 8) <= xend)) {
				xy = XYTOIND(x, y, graph->ncol);
				buff[xy] = 0xff;
				/* + 7 because we'll increment another one at the loop. */
				x += 7;
			} else {
				CLEAR_PIXEL(buff, x, y, graph->ncol);
			}
		}
	}
	
	if (redraw) {
		err = graph->display.pPixelMatrixDraw(&graph->display, graph->buff,
		                                      0, graph->ncol,
#ifdef EMWIN_WORKAROUND
		                                      graph->ncol,
#endif   /* EMWIN_WORKAROUND */
		                                      graph->y0, graph->nrow);
		RET_ON_ERR(err);
	}

	return EMSTATUS_OK;
}

/***
    GRAPHDISPLAY_GetWidth:  Get the width in pixels of the drawing area of
                            each sub-graph (all are the same).
    args:                   handle - reference to graph area
                            wdraw - width of sub-graphs
    returns:   0 if successful, > 0 on error (EMSTATUS code)
    modifies:  wgraph
***/
EMSTATUS GRAPHDISPLAY_GetWidth(GRAPHDISPLAY_Handle_t *handle, uint8_t *wdraw) {
	GRAPHDISPLAY_Device_t *graph;         /* graph data */

	graph = (GRAPHDISPLAY_Device_t *) handle;

	if ((NULL == graph) || (!graph->init)) {
		return GRAPHDISPLAY_EMSTATUS_NOTINIT;
	}

	*wdraw = graph->wdraw;

	return EMSTATUS_OK;
}


/**** Private Functions ****/

/***
    draw_frames:  Erase everything in the graph area and draw empty frames
                  for each sub-graph.  y tics are drawn on the left and right
                  sides of the first and last (no right if only one sub-graph),
                  x tics are drawn underneath.  Assumes the graph setup is
                  valid.  Only affects the internal buffer.
    args:         graph - graph data
    returns:   0 if successful, > 0 on error (EMSTATUS code)
    modifies:  graph->buff contents
***/
static EMSTATUS draw_frames(GRAPHDISPLAY_Device_t *graph) {
	uint8_t *buff;                        /* pixel buffer */
	uint8_t subID;                        /* sub-graph */
	uint8_t x, y;                         /* pixel coordinates */
	uint8_t xtic, ytic;                   /* tic coordinates */
	uint8_t xst, xend;                    /* horizontal line bounds (incl.) */
	uint8_t yst, yend;                    /* vertical line bounds (incl.) */
	uint32_t xy;                          /* pixel index in buffer */
	uint8_t i;
	uint32_t err;

	buff = (uint8_t *) graph->buff;
	/* / 8 to convert columns to bytes, ie bit-addressed.  This is going to
	   be fun ... Pixel values are also inverted - bits set to 0 show on LCD. */
	memset(buff, 255, (graph->ncol / 8) * graph->nrow);

	for (subID=0; subID<graph->ngraph; subID++) {
		/* Solid frame first. */
		for (i=1; i<=WBORDER; i++) {
			xst = X0(graph, subID) - i;
			xend = X0(graph, subID) + graph->wdraw + i - 1;
			yend = Y0(graph) + i;
			yst = yend - graph->hdraw - (2 * i);

			for (x=xst; x<=xend; x++) {
				if ((0 == (x % 8)) && ((x + 8) <= xend)) {
					xy = XYTOIND(x, yst, graph->ncol);
					buff[xy] = 0x00;
					xy = XYTOIND(x, yend, graph->ncol);
					buff[xy] = 0x00;
					/* + 7 because we'll increment another at the loop. */
					x += 7;
				} else {
					SET_PIXEL(buff, x, yst, graph->ncol);
					SET_PIXEL(buff, x, yend, graph->ncol);
				}
			}

			/* We've drawn the horizontals so can skip the first/last row. */
			yend -= 1;
			yst += 1;
			for (y=yst; y<=yend; y++) {
				SET_PIXEL(buff, xst, y, graph->ncol);
			}
			for (y=yst; y<=yend; y++) {
				SET_PIXEL(buff, xend, y, graph->ncol);
			}
		}

		/* Vertical tic marks. */
		for (x=0; x<graph->wdraw; x+=TICSEP) {
			xtic = X0(graph, subID) + x;
			/* +1 on yst because the border is outside the drawing region, ie.
			   have already used + WBORDER for the box.  -1 b/c yend inclusive. */
			yst = Y0(graph) + WBORDER + 1;
			yend = yst + TICSIZE - 1;
			for (y=yst; y<=yend; y++) {
				SET_PIXEL(buff, xtic, y, graph->ncol);
			}
		}

		/* Left tic marks only for first sub-graph, right the last (unless there's
		   only one - the else if guarantees that. */
		if (0 == subID) {
			for (y=0; y<graph->hdraw; y+=TICSEP) {
				/* -1 on xend because WBORDER is the outer box column.  +1 b/c xst
				   inclusive. */
				xend = X0(graph,subID) - WBORDER - 1;
				xst = xend - TICSIZE + 1;
				ytic = Y0(graph) - y;
				for (x=xst; x<=xend; x++) {
					SET_PIXEL(buff, x, ytic, graph->ncol);
				}
			}
		} else if ((graph->ngraph - 1) == subID) {
			for (y=0; y<graph->hdraw; y+=TICSEP) {
				/* +1 to put xst outside border, -1 b/c xend inclusive. */
				xst = X0(graph, subID) + graph->wdraw + WBORDER + 1;
				/* - 1 b/c incl. */
				xend = xst + TICSIZE - 1;
				ytic = Y0(graph) - y;
				for (x=xst; x<=xend; x++) {
					SET_PIXEL(buff, x, ytic, graph->ncol);
				}
			}
		}
	}

	return EMSTATUS_OK;
}
