
/*****
      CC_Configuration.c -
      Implementation of the ZWave Configuration class, Version 1.  It uses
      several application functions that must defined elsewhere; this file
      contains the class functions.

      c 2019-2020 Primordial Machine Vision Systems
*****/

#include <string.h>
#include <ZAF_tx_mutex.h>

#define DEBUGPRINT
#include "DebugPrint.h"

#include <CC_Configuration.h>


/**** Macros ****/

/***
    CLIP(val, min, max) - evaluate to min or max if value outside bounds,
                          otherwise to val
***/
#define CLIP(val, min, max)   \
	(((val) < (min)) ? (min) : (((val) > (max)) ? (max) : (val)))


/**** Command Implementation ****/

REGISTER_CC(COMMAND_CLASS_CONFIGURATION, CONFIGURATION_VERSION,
            handleCCConfiguration);

/***
    handleCCConfiguration:  Respond to incoming CC_CONFIGURATION commands.
    args:                   rxopt - ZWave options on received frame
                            rxcmd - incoming command
                            rxxlen - number of bytes in rxcmd
    returns:   success/failure/unsupported flag
***/
received_frame_status_t
handleCCConfiguration(RECEIVE_OPTIONS_TYPE_EX *rxopt,
                         ZW_APPLICATION_TX_BUFFER *rxcmd, uint8_t rxlen) {
	ZAF_TRANSPORT_TX_BUFFER txbuff;       /* transmit buffer */
	ZW_APPLICATION_TX_BUFFER *txcmd;      /* application data within txbuff */
	TRANSMIT_OPTIONS_TYPE_SINGLE_EX *txopt;   /* outgoing options */
	enum EQueueNotifyingStatus status;    /* transmit result */
	uint8_t txlen;                        /* length of transmit buffer */
	uint8_t paramID;                      /* parameter identifier */
	uint8_t level;                        /* packed value size (bytes) */
	int32_t val;                          /* parameter value */
	int16_t sval;                         /* 2 byte value */
	int8_t cval;                          /* 1 byte value */

	txcmd = &(txbuff.appTxBuf);
	memset(txcmd, 0, sizeof(*txcmd));
	RxToTxOptions(rxopt, &txopt);

	switch (rxcmd->ZW_Common.cmd) {
	case CONFIGURATION_GET:
		if (Check_not_legal_response_job(rxopt)) {
			DPRINTF("  CONFIG_GET not legal response!\n");
			return RECEIVED_FRAME_STATUS_FAIL;
		}
		paramID = rxcmd->ZW_ConfigurationGetFrame.parameterNumber;
		val = appConfigValue(paramID);
		level = appConfigLevel(paramID);

		switch (level & 0x07) {
		case 1:
			cval = CLIP(val, -128, 127);
			txcmd->ZW_ConfigurationReport1byteFrame.cmdClass =
				COMMAND_CLASS_CONFIGURATION;
			txcmd->ZW_ConfigurationReport1byteFrame.cmd =
				CONFIGURATION_REPORT;
			txcmd->ZW_ConfigurationReport1byteFrame.parameterNumber =
				appConfigValidParam(paramID);
			txcmd->ZW_ConfigurationReport1byteFrame.level =
				appConfigLevel(paramID);
			txcmd->ZW_ConfigurationReport1byteFrame.configurationValue1 =
				cval & 0xff;
			txlen = sizeof(ZW_CONFIGURATION_REPORT_1BYTE_FRAME);
			break;
		case 2:
			sval = CLIP(val, -32768, 32767);
			txcmd->ZW_ConfigurationReport2byteFrame.cmdClass =
				COMMAND_CLASS_CONFIGURATION;
			txcmd->ZW_ConfigurationReport2byteFrame.cmd =
				CONFIGURATION_REPORT;
			txcmd->ZW_ConfigurationReport2byteFrame.parameterNumber =
				appConfigValidParam(paramID);
			txcmd->ZW_ConfigurationReport2byteFrame.level =
				appConfigLevel(paramID);
			txcmd->ZW_ConfigurationReport2byteFrame.configurationValue1 =
				(sval >> 8) & 0x0ff;
			txcmd->ZW_ConfigurationReport2byteFrame.configurationValue2 =
				sval & 0xff;
			txlen = sizeof(ZW_CONFIGURATION_REPORT_2BYTE_FRAME);
			break;
		case 4:
			txcmd->ZW_ConfigurationReport4byteFrame.cmdClass =
				COMMAND_CLASS_CONFIGURATION;
			txcmd->ZW_ConfigurationReport4byteFrame.cmd =
				CONFIGURATION_REPORT;
			txcmd->ZW_ConfigurationReport4byteFrame.parameterNumber =
				appConfigValidParam(paramID);
			txcmd->ZW_ConfigurationReport4byteFrame.level =
				appConfigLevel(paramID);
			txcmd->ZW_ConfigurationReport4byteFrame.configurationValue1 =
				(val >> 24) & 0x0ff;
			txcmd->ZW_ConfigurationReport4byteFrame.configurationValue2 =
				(val >> 16) & 0x0ff;
			txcmd->ZW_ConfigurationReport4byteFrame.configurationValue3 =
				(val >> 8) & 0x0ff;
			txcmd->ZW_ConfigurationReport4byteFrame.configurationValue4 =
				val & 0xff;
			txlen = sizeof(ZW_CONFIGURATION_REPORT_4BYTE_FRAME);
			break;
		default:
			return RECEIVED_FRAME_STATUS_FAIL;
		}
		break;
	case CONFIGURATION_SET:
		paramID = rxcmd->ZW_ConfigurationSet1byteFrame.parameterNumber;
		level = rxcmd->ZW_ConfigurationSet1byteFrame.level;
		switch (level & 0x07) {
		case 1:
			val = rxcmd->ZW_ConfigurationSet1byteFrame.configurationValue1;
			break;
		case 2:
			val =
				(rxcmd->ZW_ConfigurationSet2byteFrame.configurationValue1 << 8) |
				rxcmd->ZW_ConfigurationSet2byteFrame.configurationValue2;
			break;
		case 4:
			val =
				(rxcmd->ZW_ConfigurationSet4byteFrame.configurationValue1 << 24) |
				(rxcmd->ZW_ConfigurationSet4byteFrame.configurationValue2 << 16) |
				(rxcmd->ZW_ConfigurationSet4byteFrame.configurationValue3 << 8) |
				rxcmd->ZW_ConfigurationSet4byteFrame.configurationValue4;
			break;
		default:
			return RECEIVED_FRAME_STATUS_FAIL;
		}

		if (appConfigSet(paramID, val, level & 0x80)) {
			return RECEIVED_FRAME_STATUS_SUCCESS;
		} else {
			return RECEIVED_FRAME_STATUS_FAIL;
		}
		break;
	case CONFIGURATION_REPORT:
		DPRINTF("  CONFIG_REPORT not supported\n");
		return RECEIVED_FRAME_STATUS_NO_SUPPORT;
	default:
		DPRINTF("  unknown CONFIGURATION command 0x%02x\n", rxcmd->ZW_Common.cmd);
		return RECEIVED_FRAME_STATUS_NO_SUPPORT;
	}

	status = Transport_SendResponseEP((uint8_t *) txcmd, txlen, txopt, NULL);
	if (EQUEUENOTIFYING_STATUS_SUCCESS == status) {
		return RECEIVED_FRAME_STATUS_SUCCESS;
	} else {
		return RECEIVED_FRAME_STATUS_FAIL;
	}
}


